/*  Misfit Model 3D
 * 
 *  Copyright (c) 2004 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 *
 *  See the COPYING file for full license text.
 */


#include "model.h"
#include "toolbox.h"
#include "tool.h"
#include "glmath.h"
#include "decalmgr.h"
#include "decal.h"
#include "log.h"
#include "statusbar.h"
#include "modelstatus.h"
#include "mviewport.h"

#include <qdragobject.h>
#include <qfont.h>
#include <stdarg.h>
#include <math.h>


static Matrix s_mat;

static unsigned const FONT_START = 32;
static unsigned const FONT_SIZE  = 96;

GLvoid glPrintf( int displayList, const char *fmt, ... )
{
    if ( fmt )
    {
       char line[1024];
       va_list ap;

       va_start( ap, fmt );
       vsprintf( line, fmt, ap );
       va_end( ap );

       glPushAttrib( GL_LIST_BIT );

       glListBase( displayList - FONT_START );

       glCallLists( strlen( line ), GL_UNSIGNED_BYTE, line );

       glPopAttrib( );
    }
}


ModelViewport::ModelViewport( QWidget * parent, const char * name )
   : QGLWidget( parent, name ),
     m_model( NULL ),
     m_viewDirection( ViewFront ),
     m_centerX( 0.0 ),
     m_centerY( 0.0 ),
     m_rotX( 0.0 ),
     m_rotY( 0.0 ),
     m_rotZ( 0.0 ),
     m_zoomLevel( 5.0 ),
     m_far( 10000.0 ),
     m_near( 1.0 ),
     m_farOrtho( 1000000.0 ),
     m_nearOrtho( 0.01 ),
     m_viewportWidth( 0 ),
     m_viewportHeight( 0 ),
     m_displayList( 0 ),
     m_display( NULL ),
     m_glFont( NULL ),
     m_texturesLoaded( false ),
     m_viewOptions( ViewTexture ),
     m_toolbox( NULL )
{
   setFocusPolicy( WheelFocus );
   setMinimumSize( 220, 180 );

   double rot[3] = { 45 * PIOVER180, 0, 0 };
   s_mat.setRotation( rot );

   m_backColor.setRgb( 130, 200, 200 );

   setAcceptDrops( true );
}

ModelViewport::~ModelViewport()
{
   if ( m_model )
   {
      m_model->removeContext( (unsigned) this );
   }

   if ( m_display && m_glFont )
   {
      XFreeFont( m_display, m_glFont );
      XCloseDisplay( m_display );
      glDeleteLists( m_displayList, FONT_SIZE );
   }
}

void ModelViewport::initializeGL()
{
   glShadeModel( GL_SMOOTH );
   glClearColor( m_backColor.red() / 256.0, 
         m_backColor.green() / 256.0, 
         m_backColor.blue() / 256.0, 1.0f );
   glClearDepth( 1.0f );
   glEnable( GL_DEPTH_TEST );
   glDepthFunc( GL_LEQUAL );
   glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

   GLfloat ambient[]  = {  0.8f,  0.8f,  0.8f,  1.0f };
   GLfloat diffuse[]  = {  1.0f,  1.0f,  1.0f,  1.0f };
   GLfloat position[] = {  0.0f,  0.0f,  1.0f,  0.0f };

   glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE );
   glLightfv( GL_LIGHT0, GL_AMBIENT, ambient );
   glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse );
   glLightfv( GL_LIGHT0, GL_POSITION, position );
   //glEnable( GL_LIGHT0 );

   {
      int texSize = 0;
      glGetIntegerv( GL_MAX_TEXTURE_SIZE, &texSize );
      log_debug( "max texture size is %dx%d\n", texSize, texSize );
   }

   checkGlErrors();

   {
#ifdef WIN32
      log_debug( "getting windows font\n" );
      QFont        fixedFont( "fixed", 8 );
      HDC glFont = qt_display_dc();

      log_debug( "creating display list\n" );
      m_displayList = glGenLists(FONT_SIZE);
      if ( m_displayList != 0 )
      {
         log_debug( "setting up font bitmap\n" );
         wglUseFontBitmaps( glFont, FONT_START, FONT_SIZE, m_displayList );
      }
#else
      log_debug( "getting linux font\n" );

      log_debug( "creating display list\n" );
      m_displayList = glGenLists(FONT_SIZE);
      if ( m_displayList != 0 )
      {
         log_debug( "setting up glx font\n" );
         m_display = XOpenDisplay( NULL );
         m_glFont  = XLoadQueryFont( m_display, "fixed" );
         glXUseXFont( m_glFont->fid, FONT_START, FONT_SIZE, m_displayList );
      }

      //fontInfo = XLoadQueryFont( dpy, "-*-Courier New-bold-i-normal-*-*-720-*-*-*-*-iso8859-1" );
#endif
   }

#ifdef MM3D_ENABLEALPHA
   glEnable( GL_BLEND );
   glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
   if ( ! glIsEnabled( GL_BLEND) )
   {
      log_warning( "alpha not supported\n" );
      glDisable( GL_BLEND );
      glGetError(); // clear errors
   }
#endif // MM3D_ENABLEALPHA
}

void ModelViewport::resizeGL( int w, int h )
{
   if ( h == 0 )
   {
      h = 1;
   }

   m_viewportWidth  = w;
   m_viewportHeight = h;

   adjustViewport();
}

void ModelViewport::paintGL()
{
   //LOG_PROFILE();

   if ( m_viewDirection == View3d )
   {
      glEnable( GL_LIGHT0 );
      glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
      glLoadIdentity( );

      glTranslatef( -m_centerX, -m_centerY, -(m_zoomLevel * 2.0) );
      glRotatef( m_rotZ, 0.0, 0.0, 1.0);
      glRotatef( m_rotY, 0.0, 1.0, 0.0);
      glRotatef( m_rotX, 1.0, 0.0, 0.0);
   }
   else
   {
      glDisable( GL_LIGHT0 );
      glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
      glLoadIdentity( );

      glTranslatef( 0.0, 0.0, -500000.0 );

      switch ( m_viewDirection )
      {
         case ViewFront:
            // do nothing
            break;
         case ViewBack:
            glRotatef( 180.0, 0.0, 1.0, 0.0);
            break;
         case ViewLeft:
            glRotatef( -90.0, 0.0, 1.0, 0.0);
            break;
         case ViewRight:
            glRotatef(  90.0, 0.0, 1.0, 0.0);
            break;
         case ViewTop:
            glRotatef(  90.0, 1.0, 0.0, 0.0);
            break;
         case ViewBottom:
            glRotatef( -90.0, 1.0, 0.0, 0.0);
            break;
         default:
            log_error( "Unknown ViewDirection: %d\n", m_viewDirection );
            swapBuffers();
            return;
            break;
      }
   }

   drawGridLines();

   if ( m_model )
   {
      glColor3f( 0.7, 0.7, 0.7 );
      if ( m_viewDirection == View3d )
      {
         glEnable( GL_LIGHTING );
         int opt = Model::DO_TEXTURE | Model::DO_SMOOTHING;
         switch ( m_viewOptions )
         {
            case ViewWireframe:
               opt = Model::DO_WIREFRAME;
               break;
            case ViewFlat:
               opt = Model::DO_NONE;
               break;
            case ViewSmooth:
               opt = Model::DO_SMOOTHING;
               break;
            default:
               break;
         }
         m_model->draw( opt, (unsigned) this );
      }
      else
      {
         glClear( GL_DEPTH_BUFFER_BIT );
         m_model->drawLines();
         m_model->drawPoints();
         glDisable( GL_DEPTH_TEST );
         m_model->drawJoints();
         for ( DecalList::iterator it = m_decals.begin(); it != m_decals.end(); it++ )
         {
            (*it)->draw();
         }
         glEnable( GL_DEPTH_TEST );
      }
   }
   else
   {
      glBegin( GL_QUADS );

      glColor3f( 1.0, 0.0, 0.0 );     // front

      glVertex3f( -1.0, -1.0,  1.0 );
      glVertex3f( -1.0,  1.0,  1.0 );
      glVertex3f(  1.0,  1.0,  1.0 );
      glVertex3f(  1.0, -1.0,  1.0 );

      glColor3f( 0.0, 1.0, 1.0 );     // back

      glVertex3f( -1.0, -1.0, -1.0 );
      glVertex3f( -1.0,  1.0, -1.0 );
      glVertex3f(  1.0,  1.0, -1.0 );
      glVertex3f(  1.0, -1.0, -1.0 );

      glColor3f( 0.0, 0.0, 1.0 );     // top

      glVertex3f(  1.0,  1.0,  1.0 );
      glVertex3f(  1.0,  1.0, -1.0 );
      glVertex3f( -1.0,  1.0, -1.0 );
      glVertex3f( -1.0,  1.0,  1.0 );

      glColor3f( 1.0, 1.0, 0.0 );     // bottom

      glVertex3f(  1.0, -1.0,  1.0 );
      glVertex3f(  1.0, -1.0, -1.0 );
      glVertex3f( -1.0, -1.0, -1.0 );
      glVertex3f( -1.0, -1.0,  1.0 );

      glColor3f( 0.0, 1.0, 0.0 );     // left

      glVertex3f( -1.0,  1.0,  1.0 );
      glVertex3f( -1.0,  1.0, -1.0 );
      glVertex3f( -1.0, -1.0, -1.0 );
      glVertex3f( -1.0, -1.0,  1.0 );

      glColor3f( 1.0, 0.0, 1.0 );     // right

      glVertex3f(  1.0,  1.0,  1.0 );
      glVertex3f(  1.0,  1.0, -1.0 );
      glVertex3f(  1.0, -1.0, -1.0 );
      glVertex3f(  1.0, -1.0,  1.0 );

      glEnd();
   }

   glDisable( GL_LIGHTING );

   drawOrigin();

   checkGlErrors();

   swapBuffers();
}

void ModelViewport::drawGridLines()
{
   if ( m_viewDirection == View3d )
   {
      double max = 25.0;
      double inc = 5.0;
      double x;
      double z;

      glColor3f( 0.55f, 0.55f, 0.55f );

      glBegin( GL_LINES );

      for ( x = -max; x <= max; x += inc )
      {
         glVertex3f( x, 0, -max );
         glVertex3f( x, 0, +max );
      }

      for ( z = -max; z <= max; z += inc )
      {
         glVertex3f( -max, 0, z );
         glVertex3f( +max, 0, z );
      }

      glEnd();
   }
   else
   {
      double unitWidth = 1.0;
      double maxDimension = (m_width > m_height) ? m_width : m_height;

      while ( (maxDimension / unitWidth) > 8 )
      {
         unitWidth *= 4.0;
      }
      while ( (maxDimension / unitWidth) < 2 )
      {
         unitWidth /= 4.0;
      }

      double xRangeMin = 0;
      double xRangeMax = 0;
      double yRangeMin = 0;
      double yRangeMax = 0;

      double xStart = 0;
      double yStart = 0;

      double x = 0;
      double y = 0;

      glColor3f( 0.55f, 0.55f, 0.55f );

      glBegin( GL_LINES );
      
      switch ( m_viewDirection )
      {
         case ViewFront:
            xRangeMin = m_centerX - (m_width / 2.0);
            xRangeMax = m_centerX + (m_width / 2.0);
            yRangeMin = m_centerY - (m_height / 2.0);
            yRangeMax = m_centerY + (m_height / 2.0);

            xStart = unitWidth * ((int) (xRangeMin / unitWidth));
            yStart = unitWidth * ((int) (yRangeMin / unitWidth));

            for ( x = xStart; x < xRangeMax; x += unitWidth )
            {
               glVertex3f( x, yRangeMin, 0 );
               glVertex3f( x, yRangeMax, 0 );
            }

            for ( y = yStart; y < yRangeMax; y += unitWidth )
            {
               glVertex3f( xRangeMin, y, 0 );
               glVertex3f( xRangeMax, y, 0 );
            }
            break;

         case ViewBack:
            xRangeMin = -m_centerX - (m_width / 2.0);
            xRangeMax = -m_centerX + (m_width / 2.0);
            yRangeMin = m_centerY - (m_height / 2.0);
            yRangeMax = m_centerY + (m_height / 2.0);

            xStart = unitWidth * ((int) (xRangeMin / unitWidth));
            yStart = unitWidth * ((int) (yRangeMin / unitWidth));

            for ( x = xStart; x < xRangeMax; x += unitWidth )
            {
               glVertex3f( x, yRangeMin, 0 );
               glVertex3f( x, yRangeMax, 0 );
            }

            for ( y = yStart; y < yRangeMax; y += unitWidth )
            {
               glVertex3f( xRangeMin, y, 0 );
               glVertex3f( xRangeMax, y, 0 );
            }
            break;

         case ViewLeft:
            xRangeMin = -m_centerX - (m_width / 2.0);
            xRangeMax = -m_centerX + (m_width / 2.0);
            yRangeMin = m_centerY - (m_height / 2.0);
            yRangeMax = m_centerY + (m_height / 2.0);

            xStart = unitWidth * ((int) (xRangeMin / unitWidth));
            yStart = unitWidth * ((int) (yRangeMin / unitWidth));

            for ( x = xStart; x < xRangeMax; x += unitWidth )
            {
               glVertex3f( 0, yRangeMin, x );
               glVertex3f( 0, yRangeMax, x );
            }

            for ( y = yStart; y < yRangeMax; y += unitWidth )
            {
               glVertex3f( 0, y, xRangeMin );
               glVertex3f( 0, y, xRangeMax );
            }
            break;

         case ViewRight:
            xRangeMin = m_centerX - (m_width / 2.0);
            xRangeMax = m_centerX + (m_width / 2.0);
            yRangeMin = m_centerY - (m_height / 2.0);
            yRangeMax = m_centerY + (m_height / 2.0);

            xStart = unitWidth * ((int) (xRangeMin / unitWidth));
            yStart = unitWidth * ((int) (yRangeMin / unitWidth));

            for ( x = xStart; x < xRangeMax; x += unitWidth )
            {
               glVertex3f( 0, yRangeMin, x );
               glVertex3f( 0, yRangeMax, x );
            }

            for ( y = yStart; y < yRangeMax; y += unitWidth )
            {
               glVertex3f( 0, y, xRangeMin );
               glVertex3f( 0, y, xRangeMax );
            }
            break;

         case ViewTop:
            xRangeMin = m_centerX - (m_width / 2.0);
            xRangeMax = m_centerX + (m_width / 2.0);
            yRangeMin = -m_centerY - (m_height / 2.0);
            yRangeMax = -m_centerY + (m_height / 2.0);

            xStart = unitWidth * ((int) (xRangeMin / unitWidth));
            yStart = unitWidth * ((int) (yRangeMin / unitWidth));

            for ( x = xStart; x < xRangeMax; x += unitWidth )
            {
               glVertex3f( x, 0, yRangeMin );
               glVertex3f( x, 0, yRangeMax );
            }

            for ( y = yStart; y < yRangeMax; y += unitWidth )
            {
               glVertex3f( xRangeMin, 0, y );
               glVertex3f( xRangeMax, 0, y );
            }
            break;

         case ViewBottom:
            xRangeMin = m_centerX - (m_width / 2.0);
            xRangeMax = m_centerX + (m_width / 2.0);
            yRangeMin = m_centerY - (m_height / 2.0);
            yRangeMax = m_centerY + (m_height / 2.0);

            xStart = unitWidth * ((int) (xRangeMin / unitWidth));
            yStart = unitWidth * ((int) (yRangeMin / unitWidth));

            for ( x = xStart; x < xRangeMax; x += unitWidth )
            {
               glVertex3f( x, 0, yRangeMin );
               glVertex3f( x, 0, yRangeMax );
            }

            for ( y = yStart; y < yRangeMax; y += unitWidth )
            {
               glVertex3f( xRangeMin, 0, y );
               glVertex3f( xRangeMax, 0, y );
            }
            break;

         default:
            log_error( "Unhandled view direction: %d\n", m_viewDirection );
            break;
      }

      glEnd();

      glColor3f( 0.35f, 0.35f, 0.35f );
      glRasterPos3f( xRangeMin, yRangeMin, 0.0f );
      switch ( m_viewDirection )
      {
         case ViewFront:
            break;

         case ViewBack:
            glRasterPos3f( xRangeMax, yRangeMin, 0.0f );
            break;

         case ViewLeft:
            glRasterPos3f( 0.0f, yRangeMin, xRangeMax );
            break;

         case ViewRight:
            glRasterPos3f( 0.0f, yRangeMin, xRangeMin );
            break;

         case ViewTop:
            glRasterPos3f( xRangeMin, 0.0f, yRangeMax );
            break;

         case ViewBottom:
            glRasterPos3f( xRangeMin, 0.0f, yRangeMin );
            break;

         default:
            log_error( "Unhandled view direction: %d\n", m_viewDirection );
            break;
      }
      glPrintf( m_displayList, "%g", unitWidth );
   }
}

void ModelViewport::drawOrigin()
{
   glDisable( GL_DEPTH_TEST );

   glBegin( GL_LINES );

   glColor3f( 1, 0, 0 );
   glVertex3f( 0, 0, 0 );
   glVertex3f( m_zoomLevel / 10, 0, 0 );
   glColor3f( 0, 1, 0 );
   glVertex3f( 0, 0, 0 );
   glVertex3f( 0, m_zoomLevel / 10, 0 );
   glColor3f( 0, 0, 1 );
   glVertex3f( 0, 0, 0 );
   glVertex3f( 0, 0, m_zoomLevel / 10 );

   glEnd();

   glEnable( GL_DEPTH_TEST );
}

void ModelViewport::adjustViewport()
{
   /* Setup our viewport. */
   glViewport( 0, 0, ( GLint ) m_viewportWidth, ( GLint ) m_viewportHeight );

   /* change to the projection matrix and set our viewing volume. */
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity( );

   /* Set our perspective */
   GLfloat ratio = ( GLfloat ) m_viewportWidth / ( GLfloat ) m_viewportHeight;
   if ( m_viewDirection == View3d )
   {
      m_far  = m_zoomLevel * 2000.0;
      m_near = m_zoomLevel * 0.002;
      gluPerspective( 45.0f, ratio, m_near, m_far );

      float x = 0.0;
      float y = 0.0;
      if ( m_viewportHeight > m_viewportWidth )
      {
         x = m_zoomLevel;
         y = x / ratio;
      }
      else
      {
         y = m_zoomLevel;
         x = y * ratio;
      }

      m_width = x * 2.0;
      m_height = y * 2.0;
   }
   else
   {
      float x = 0.0;
      float y = 0.0;
      if ( m_viewportHeight > m_viewportWidth )
      {
         x = m_zoomLevel;
         y = x / ratio;
      }
      else
      {
         y = m_zoomLevel;
         x = y * ratio;
      }
      glOrtho( m_centerX - x, m_centerX + x, 
            m_centerY - y, m_centerY + y, 
            m_nearOrtho, m_farOrtho );

      m_width = x * 2.0;
      m_height = y * 2.0;
   }

   /* Make sure we're changing the model view and not the projection */
   glMatrixMode( GL_MODELVIEW );
   glLoadIdentity( );

   updateGL();
}

void ModelViewport::wheelEvent( QWheelEvent * e )
{
   if ( e->delta() > 0 )
   {
      zoomIn();
   }
   else
   {
      zoomOut();
   }
}

void ModelViewport::zoomIn()
{
   m_zoomLevel *= (0.75);

   QString zoomStr;
   zoomStr.sprintf( "%f", m_zoomLevel );
   emit zoomLevelChanged( zoomStr );

   makeCurrent();
   adjustViewport();
}

void ModelViewport::zoomOut()
{
   if ( m_zoomLevel * (4.0 / 3.0 ) < 250000 )
   {
      m_zoomLevel *= (4.0 / 3.0);  // as accurate as possible for zoom out
   }

   QString zoomStr;
   zoomStr.sprintf( "%f", m_zoomLevel );
   emit zoomLevelChanged( zoomStr );

   makeCurrent();
   adjustViewport();
}

static int tempButton = 0;

void ModelViewport::mousePressEvent( QMouseEvent * e )
{
   //printf( "press = %d\n", e->button() );
   e->accept();
   tempButton = e->button();
   if ( e->button() == MidButton || (m_viewDirection == View3d && e->button() == LeftButton) )
   {
      m_scrollStartPosition = e->pos();
   }
   else
   {
      int button = constructButtonState( e );

      Tool * tool = m_toolbox->getCurrentTool();
      tool->mouseButtonDown( this, button, e->pos().x(), e->pos().y() );
   }
}

void ModelViewport::mouseReleaseEvent( QMouseEvent * e )
{
   //printf( "release = %d\n", e->button() );
   e->accept();
   if ( tempButton == MidButton || (m_viewDirection == View3d && tempButton == LeftButton) )
   {
      // do nothing
   }
   else
   {
      int button = constructButtonState( e );

      Tool * tool = m_toolbox->getCurrentTool();
      tool->mouseButtonUp( this, button, e->pos().x(), e->pos().y() );
      m_model->operationComplete( tool->getName( 0 ) );
   }
}

void ModelViewport::mouseMoveEvent( QMouseEvent * e )
{
   e->accept();
   if ( tempButton == MidButton || (m_viewDirection == View3d && tempButton == LeftButton) )
   {
      if ( m_viewDirection == View3d && (tempButton == LeftButton) )
      {
#if 1
         QPoint curPos = e->pos();

         double xDiff = -(m_scrollStartPosition.x() - curPos.x());
         double yDiff = -(m_scrollStartPosition.y() - curPos.y());

         Matrix mcur;
         Matrix mcurinv;
         double rot[3];
         rot[0] = m_rotX * PIOVER180;
         rot[1] = m_rotY * PIOVER180;
         rot[2] = m_rotZ * PIOVER180;
         mcur.setRotation( rot );
         mcurinv = mcur.getInverse();

         Vector yvec;
         yvec.set( 0, 0.0 );
         yvec.set( 1, 1.0 );
         yvec.set( 2, 0.0 );
         yvec.set( 3, 0.0 );

         Vector xvec;
         xvec.set( 0, 1.0 );
         xvec.set( 1, 0.0 );
         xvec.set( 2, 0.0 );
         xvec.set( 3, 0.0 );

         Matrix mx;
         Matrix my;
         double rotY = (double) xDiff / (double) this->width()  * 3.14159 * 2.0;
         double rotX = (double) yDiff / (double) this->height() * 3.14159 * 2.0;

         yvec = yvec * mcurinv;
         my.setRotationOnAxis( yvec.getVector(), rotY );

         xvec = xvec * mcurinv;
         mx.setRotationOnAxis( xvec.getVector(), rotX );

         mcur = mx * mcur;
         mcur = my * mcur;
         mcur.getRotation( rot );
         m_rotX = rot[0] / PIOVER180;
         m_rotY = rot[1] / PIOVER180;
         m_rotZ = rot[2] / PIOVER180;

         //m_rotY += xDiff * PIOVER180 * 14.0;
         //m_rotX += yDiff * PIOVER180 * 14.0;

         m_scrollStartPosition = curPos;
#else
         QPoint curPos = e->pos();

         double xDiff = -(m_scrollStartPosition.x() - curPos.x());
         double yDiff = -(m_scrollStartPosition.y() - curPos.y());

         m_rotY += xDiff * PIOVER180 * 14.0;
         m_rotX += yDiff * PIOVER180 * 14.0;

         m_scrollStartPosition = curPos;
#endif

         // And finally, update the view
         makeCurrent();
         adjustViewport();
      }
      else
      {
         //printf( "adjusting translation\n" );
         QPoint curPos = e->pos();

         double xDiff = m_scrollStartPosition.x() - curPos.x();
         m_centerX += xDiff * (m_width / m_viewportWidth);

         double yDiff = m_scrollStartPosition.y() - curPos.y();
         yDiff = -yDiff;  // Adjust for difference between pixel and GL coordinates
         m_centerY += yDiff * (m_height / m_viewportHeight);

         m_scrollStartPosition = curPos;

         makeCurrent();
         adjustViewport();
      }
   }
   else
   {
      //printf( "tool mouse event\n" );
      int button = constructButtonState( e );

      Tool * tool = m_toolbox->getCurrentTool();
      tool->mouseButtonMove( this, button, e->pos().x(), e->pos().y() );
   }
}

void ModelViewport::keyPressEvent( QKeyEvent * e )
{
   switch ( e->key() )
   {
      case Key_Equal:
      case Key_Plus:
         zoomIn();
         break;
      case Key_Minus:
      case Key_Underscore:
         zoomOut();
         break;
      case Key_QuoteLeft:
         {
            int newDir = 0;
            switch ( m_viewDirection )
            {
               case ViewFront:
                  newDir = ViewBack;
                  break;
               case ViewBack:
                  newDir = ViewFront;
                  break;
               case ViewRight:
                  newDir = ViewLeft;
                  break;
               case ViewLeft:
                  newDir = ViewRight;
                  break;
               case ViewTop:
                  newDir = ViewBottom;
                  break;
               case ViewBottom:
                  newDir = ViewTop;
                  break;
               default:
                  break;
            }

            emit viewDirectionChanged( newDir );
            //viewChangeEvent( newDir );
         }
         break;
      case Key_0:
         m_centerX = 0.0;
         m_centerY = 0.0;
         makeCurrent();
         adjustViewport();
         break;
      default:
         QGLWidget::keyPressEvent( e );
         break;
   }
}

void ModelViewport::viewChangeEvent( int dir )
{
   float x = 0.0;
   float y = 0.0;
   float z = 0.0;

   switch ( m_viewDirection )
   {
      case ViewFront:
         x = m_centerX;
         y = m_centerY;
         break;
      case ViewBack:
         x = -m_centerX;
         y = m_centerY;
         break;
      case ViewRight:
         z = -m_centerX;
         y = m_centerY;
         break;
      case ViewLeft:
         z = m_centerX;
         y = m_centerY;
         break;
      case ViewTop:
         x = m_centerX;
         z = m_centerY;
         break;
      case ViewBottom:
         x = m_centerX;
         z = -m_centerY;
         break;
      default:
         break;
   }

   switch ( dir )
   {
      case ViewFront:
         m_centerX = x;
         m_centerY = y;
         break;
      case ViewBack:
         m_centerX = -x;
         m_centerY = y;
         break;
      case ViewRight:
         m_centerX = -z;
         m_centerY = y;
         break;
      case ViewLeft:
         m_centerX = z;
         m_centerY = y;
         break;
      case ViewTop:
         m_centerX = x;
         m_centerY = z;
         break;
      case ViewBottom:
         m_centerX = x;
         m_centerY = -z;
         break;
      default:
         break;
   }

   m_viewDirection = (ViewDirection) dir;
   makeCurrent();
   adjustViewport();
}

void ModelViewport::setZoomLevel( double zoom )
{
   m_zoomLevel = zoom;
   makeCurrent();
   adjustViewport();

   QString zoomStr;
   zoomStr.sprintf( "%f", m_zoomLevel );
   emit zoomLevelChanged( zoomStr );
}

void ModelViewport::wireframeEvent()
{
   m_viewOptions = ViewWireframe;
   updateGL();
}

void ModelViewport::flatEvent()
{
   m_viewOptions = ViewFlat;
   updateGL();
}

void ModelViewport::smoothEvent()
{
   m_viewOptions = ViewSmooth;
   updateGL();
}

void ModelViewport::textureEvent()
{
   m_viewOptions = ViewTexture;
   updateGL();
}

void ModelViewport::focusInEvent( QFocusEvent * e )
{
   m_backColor.setRgb( 150, 210, 210 );
   makeCurrent();
   glClearColor( m_backColor.red() / 256.0, 
         m_backColor.green() / 256.0, 
         m_backColor.blue() / 256.0, 1.0f );
   updateGL();
}

void ModelViewport::focusOutEvent( QFocusEvent * e )
{
   m_backColor.setRgb( 130, 200, 200 );
   makeCurrent();
   glClearColor( m_backColor.red() / 256.0, 
         m_backColor.green() / 256.0, 
         m_backColor.blue() / 256.0, 1.0f );
   updateGL();
}

void ModelViewport::dragEnterEvent( QDragMoveEvent * e )
{
   log_debug( "got drag enter event\n" );
   if ( QUriDrag::canDecode( e ) )
   {
      log_debug( "is URI\n" );
   }
   if ( QTextDrag::canDecode( e ) )
   {
      log_debug( "is Text\n" );
   }
   if ( QImageDrag::canDecode( e ) )
   {
      log_debug( "is Image\n" );
   }
   e->accept( QRect( 0, 0, this->width(), this->height() ) );
}

bool ModelViewport::getXValue( int x, int y, double * val )
{
   switch ( m_viewDirection )
   {
      case ViewFront:
      case ViewTop:
      case ViewBottom:
         *val = (m_centerX - (m_width / 2.0)) + (((double) x / (double) m_viewportWidth) * m_width);
         break;
      case ViewBack:
         *val = ((-m_centerX) + (m_width / 2.0)) + (((double) (-x) / (double) m_viewportWidth) * m_width);
         break;
      case ViewRight:
      case ViewLeft:
      default:
         return false;
   }

   return true;
}

bool ModelViewport::getYValue( int x, int y, double * val )
{
   switch ( m_viewDirection )
   {
      case ViewFront:
      case ViewBack:
      case ViewRight:
      case ViewLeft:
         *val = (m_centerY + (m_height / 2.0)) - (((double) y / (double) m_viewportHeight) * m_height);
         break;
      case ViewTop:
      case ViewBottom:
      default:
         return false;
         break;
   }

   return true;
}

bool ModelViewport::getZValue( int x, int y, double * val )
{
   switch ( m_viewDirection )
   {
      case ViewTop:
         *val = ((-m_centerY) - (m_height / 2.0)) - (((double) (-y) / (double) m_viewportHeight) * m_height);
         break;
      case ViewBottom:
         *val = (m_centerY + (m_height / 2.0)) - (((double) y / (double) m_viewportHeight) * m_height);
         //*val = (m_centerY + (m_height / 2.0)) - (((double) y / (double) m_viewportHeight) * m_height);
         break;
      case ViewRight:
         *val = (m_centerX - (m_width / 2.0)) + (((double) x / (double) m_viewportWidth) * m_width);
         break;
      case ViewLeft:
         *val = ((-m_centerX) + (m_width / 2.0)) + (((double) (-x) / (double) m_viewportWidth) * m_width);
         break;
      case ViewFront:
      case ViewBack:
      default:
         return false;
   }

   return true;
}

int ModelViewport::constructButtonState( QMouseEvent * e )
{
   int button = 0;

   //switch ( e->button() )
   switch ( tempButton )
   {
      case LeftButton:
         button = Tool::BS_Left;
         break;
      case MidButton:
         button = Tool::BS_Middle;
         break;
      case RightButton:
         button = Tool::BS_Right;
         break;
      default:
         break;
   }

   if ( e->state() & ShiftButton )
   {
      button |= Tool::BS_Shift;
   }

   if ( e->state() & AltButton )
   {
      button |= Tool::BS_Alt;
   }

   if ( e->state() & ControlButton )
   {
      button |= Tool::BS_Ctrl;
   }

   return button;
}

void ModelViewport::updateView()
{
   updateGL();
   StatusBar * bar = StatusBar::getStatusBarFromModel( m_model );
   bar->setVertices(   m_model->getVertexCount() );
   bar->setFaces(      m_model->getTriangleCount() );
   bar->setGroups(     m_model->getGroupCount() );
   bar->setBoneJoints( m_model->getBoneJointCount() );
   bar->setTextures(   m_model->getTextureCount() );
}

void ModelViewport::update3dView()
{
   if ( m_viewDirection == View3d )
   {
      updateView();
   }
}

void ModelViewport::addDecal( Decal * decal )
{
   m_decals.push_back( decal );
   updateGL();
}

void ModelViewport::removeDecal( Decal * decal )
{
   m_decals.remove( decal );
   updateGL();
}

void ModelViewport::frameArea( double x1, double y1, double z1, double x2, double y2, double z2 )
{
   double centerX = (x1 + x2) / 2.0;
   double centerY = (y1 + y2) / 2.0;
   double centerZ = (z1 + z2) / 2.0;

   double width  = fabs( x1 - x2 );
   double height = fabs( y1 - y2 );
   double depth  = fabs( z1 - z2 );

   if ( width < 0.001 )
   {
      width = 0.001;
   }
   if ( height < 0.001 )
   {
      height = 0.001;
   }
   if ( depth < 0.001 )
   {
      depth = 0.001;
   }

   width  *= 1.20;
   height *= 1.20;
   depth  *= 1.20;

   double viewWidth   = 0.0;
   double viewHeight  = 0.0;

   switch ( m_viewDirection )
   {
      case ViewFront:
         m_centerX   = centerX;
         m_centerY   = centerY;
         viewWidth   = width;
         viewHeight  = height;
         break;
      case ViewBack:
         m_centerX   = -centerX;
         m_centerY   = centerY;
         viewWidth   = width;
         viewHeight  = height;
         break;
      case ViewRight:
         m_centerX   = centerZ;
         m_centerY   = centerY;
         viewWidth   = depth;
         viewHeight  = height;
         break;
      case ViewLeft:
         m_centerX   = -centerZ;
         m_centerY   = centerY;
         viewWidth   = depth;
         viewHeight  = height;
         break;
      case ViewTop:
         m_centerX   = centerX;
         m_centerY   = -centerZ;
         viewWidth   = width;
         viewHeight  = depth;
         break;
      case ViewBottom:
         m_centerX   = centerX;
         m_centerY   = centerZ;
         viewWidth   = width;
         viewHeight  = depth;
         break;
      case View3d:
         {
            double rotation[3];
            rotation[0] = m_rotX * PIOVER180;
            rotation[1] = m_rotY * PIOVER180;
            rotation[2] = m_rotZ * PIOVER180;
            Matrix m;
            m.setRotation( rotation );

            double v[3];
            v[0] = centerX;
            v[1] = centerY;
            v[2] = centerZ;

            m.inverseRotateVector( v );

            m_centerX   = v[0];
            m_centerY   = v[1];
            viewWidth   = height;
            viewHeight  = width;
         }
         break;
      default:
         return;
         break;
   }

   if ( viewWidth > viewHeight )
   {
      if ( m_viewDirection == View3d )
      {
         m_zoomLevel = viewWidth;
      }
      else
      {
         m_zoomLevel = viewWidth / 2.0;
      }
   }
   else
   {
      if ( m_viewDirection == View3d )
      {
         m_zoomLevel = viewHeight;
      }
      else
      {
         m_zoomLevel = viewHeight / 2.0;
      }
   }

   QString zoomStr;
   zoomStr.sprintf( "%f", m_zoomLevel );
   emit zoomLevelChanged( zoomStr );

   makeCurrent();
   adjustViewport();
}

void ModelViewport::checkGlErrors()
{
   int error = glGetError();
   if ( error )
   {
      switch ( error )
      {
         case GL_INVALID_VALUE:
            model_status( m_model, StatusNormal, STATUSTIME_NONE, "OpenGL error = Invalid Value" );
            break;
         case GL_INVALID_ENUM:
            model_status( m_model, StatusNormal, STATUSTIME_NONE, "OpenGL error = Invalid Enum" );
            break;
         case GL_INVALID_OPERATION:
            model_status( m_model, StatusNormal, STATUSTIME_NONE, "OpenGL error = Invalid Operation" );
            break;
         case GL_STACK_OVERFLOW:
            model_status( m_model, StatusNormal, STATUSTIME_NONE, "OpenGL error = Stack Overflow" );
            break;
         case GL_STACK_UNDERFLOW:
            model_status( m_model, StatusNormal, STATUSTIME_NONE, "OpenGL error = Stack Underflow" );
            break;
         case GL_OUT_OF_MEMORY:
            model_status( m_model, StatusNormal, STATUSTIME_NONE, "OpenGL error = Out Of Memory" );
            break;
         default:
            model_status( m_model, StatusNormal, STATUSTIME_NONE, "OpenGL error = %d", error );
            break;
      }
   }
}
